import datetime
import os
from abc import ABC, abstractmethod
from typing import Any, Dict, List, Optional

from jinja2 import Template

from jinja_tree.app import dump
from jinja_tree.app.config import Config


class ContextPort(ABC):
    """This is the abstract interface for ContextPort adapters."""

    @abstractmethod
    def __init__(self, config: Config, plugin_config: Dict[str, Any]):
        """
        Construct a new ContextPort object given a configuration object
        and a plugin configuration dict.

        Args:
            config (Config): The configuration object.
            plugin_config (Dict[str, Any]): The plugin configuration dict.

        """
        pass

    @abstractmethod
    def get_context(self) -> Dict[str, Any]:
        """
        Retrieve the Jinja context to apply.

        Note: it can depends on the current working directory (CWD).

        Returns:
            The context dictionary.

        """
        pass

    @classmethod
    @abstractmethod
    def get_config_name(cls) -> str:
        """
        Return the name of the configuration object.

        For example, if we return "foo", it means that the configuration in the TOML file
        is located under:

        [context.foo]
        # ... some configuration ...

        Returns:
            The name of the configuration object.

        """
        pass


def make_generated_comment(
    template_string: str,
    utcnow: str,
    absolute_path: str,
    dirname: str,
    basename: str,
    relative_filepath: str,
) -> str:
    return Template(source=template_string).render(
        utcnow=utcnow,
        absolute_path=absolute_path,
        dirname=dirname,
        basename=basename,
        relative_filepath=relative_filepath,
    )


class ContextService:
    """This service is responsible for retrieving the Jinja context to apply.

    It's a kind of proxy object on ContextPort adapters to factorize some common logic.

    Attributes:
        config: The configuration object.
        adapters: The ContextPort adapters to use.

    """

    def __init__(self, config: Config, adapters: List[ContextPort]):
        self.config = config
        self.adapters = adapters
        self.comment_line1_template = self.config.context_generated_comment_line1
        self.comment_line2_template = self.config.context_generated_comment_line2
        if self.config.verbose:
            dump("initial context", self.get_context())

    def add_extra_keys_to_context(
        self, context: Dict[str, Any], absolute_path: Optional[str] = None
    ):
        context["JINJA_TREE"] = "1"
        try:
            utcnow = datetime.datetime.now(datetime.UTC).isoformat()[0:19] + "Z"  # type: ignore
        except AttributeError:
            # for python <= 3.10
            utcnow = datetime.datetime.utcnow().isoformat()[0:19] + "Z"  # type: ignore

        context["JINJA_DT"] = utcnow
        if absolute_path:
            context["JINJA_TREE_FILEPATH"] = absolute_path
            dirname = os.path.dirname(absolute_path)
            context["JINJA_TREE_DIRNAME"] = dirname
            basename = os.path.basename(absolute_path)
            context["JINJA_TREE_BASENAME"] = basename
            context["JINJA_TREE_ROOT_DIR"] = self.config.root_dir
            if absolute_path.startswith(self.config.root_dir):
                # should be true...
                relative_filepath = absolute_path[len(self.config.root_dir) + 1 :]
                context["JINJA_TREE_RELATIVE_FILEPATH"] = relative_filepath
                make_generated_comment_kwargs = {
                    "utcnow": utcnow,
                    "absolute_path": absolute_path,
                    "dirname": dirname,
                    "basename": basename,
                    "relative_filepath": relative_filepath,
                }
                comment_line1 = make_generated_comment(
                    self.comment_line1_template, **make_generated_comment_kwargs
                )
                comment_line2 = make_generated_comment(
                    self.comment_line2_template, **make_generated_comment_kwargs
                )
                context["JINJA_TREE_STYLE1_GENERATED_COMMENT"] = "\n".join(
                    [f"<!-- {x} -->" for x in (comment_line1, comment_line2) if x]
                )
                context["JINJA_TREE_STYLE2_GENERATED_COMMENT"] = "\n".join(
                    [f"# {x}" for x in (comment_line1, comment_line2) if x]
                )
                context["JINJA_TREE_STYLE3_GENERATED_COMMENT"] = "\n".join(
                    [f"// {x}" for x in (comment_line1, comment_line2) if x]
                )

    def get_context(self, absolute_path: Optional[str] = None) -> Dict[str, Any]:
        """
        Retrieve the Jinja context to apply.

        Note: it can depends on the current working directory (CWD).

        Returns:
            The context dictionary.

        """
        res: Dict[str, Any] = {}
        for adapter in self.adapters:
            res = {**res, **adapter.get_context()}
        self.add_extra_keys_to_context(res, absolute_path=absolute_path)
        return res



import os
from typing import Any, Dict, Optional

import stlog

from jinja_tree.app.action import (
    ActionPort,
    BrowseDirectoryAction,
    DirectoryAction,
    FileAction,
    IgnoreDirectoryAction,
    IgnoreFileAction,
    ProcessFileAction,
)
from jinja_tree.app.config import Config
from jinja_tree.infra.utils import is_fnmatch_ignored

FILENAME_IGNORES_DEFAULT = [".*"]
DIRNAME_IGNORES_DEFAULT = [
    "venv",
    "site-packages",
    "__pypackages__",
    "node_modules",
    "__pycache__",
    ".*",
]

DEFAULT_EXTENSIONS = [".template"]
REPLACE_DEFAULT = True
DELETE_ORIGINAL_DEFAULT = False

logger = stlog.getLogger("jinja-tree")


class ExtensionsActionAdapter(ActionPort):
    def __init__(self, config: Config, plugin_config: Dict[str, Any]):
        self.config = config
        self.plugin_config = plugin_config
        self.extensions = plugin_config.get("extensions", DEFAULT_EXTENSIONS)
        self.filename_ignores = plugin_config.get(
            "filename_ignores", FILENAME_IGNORES_DEFAULT
        )
        self.dirname_ignores = plugin_config.get(
            "dirname_ignores", DIRNAME_IGNORES_DEFAULT
        )
        self.replace = plugin_config.get("replace", REPLACE_DEFAULT)
        self.delete_original = plugin_config.get(
            "delete_original", DELETE_ORIGINAL_DEFAULT
        )

    @classmethod
    def get_config_name(self) -> str:
        return "extension"

    def trace(self, msg: str, **kwargs):
        if self.config.verbose:
            logger.debug(msg, **kwargs)

    def get_file_action(self, absolute_path: str) -> FileAction:
        if is_fnmatch_ignored(os.path.basename(absolute_path), self.filename_ignores):
            self.trace(
                "Ignored file because of ignores configuration value",
                path=absolute_path,
                filename_ignores=self.filename_ignores,
            )
            return IgnoreFileAction(source_absolute_path=absolute_path)
        target_absolute_path: Optional[str] = None
        for extension in self.extensions:
            if absolute_path.endswith(extension):
                target_absolute_path = absolute_path[0 : -(len(extension))]
        if target_absolute_path is None:
            # break not encountered
            self.trace(
                "Ignored file because of its extension",
                path=absolute_path,
                extensions=self.extensions,
            )
            return IgnoreFileAction(source_absolute_path=absolute_path)
        if os.path.exists(target_absolute_path) and not self.replace:
            logger.warning(
                f"target file: {target_absolute_path} already exists and replace config parameter is False => ignoring"
            )
            return IgnoreFileAction(source_absolute_path=absolute_path)
        return ProcessFileAction(
            source_absolute_path=absolute_path,
            target_absolute_path=target_absolute_path,
            delete_original=self.delete_original,
        )

    def get_directory_action(self, absolute_path: str) -> DirectoryAction:
        if is_fnmatch_ignored(os.path.basename(absolute_path), self.dirname_ignores):
            self.trace(
                "Ignored directory because of dirname_ignores configuration value",
                path=absolute_path,
                dirname_ignores=self.dirname_ignores,
            )
            return IgnoreDirectoryAction(source_absolute_path=absolute_path)
        return BrowseDirectoryAction(source_absolute_path=absolute_path)



from urllib.parse import quote

from jinja2 import pass_eval_context as eval_context
from jinja2.ext import Extension


@eval_context
def urlencode_filter(eval_ctx, value, safe="/"):
    return quote(value, safe=safe)


class UrlEncodeExtension(Extension):
    """Jinja2 extension to urlencode a string.

    Example:
    {{ 'https://google.com'|urlencode(safe='') }}

    => "https%3A%2F%2Fgoogle.com"

    """

    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["urlencode"] = urlencode_filter



from pathlib import Path
from typing import List, Optional

try:
    from typing import Annotated
except ImportError:
    from typing_extensions import Annotated  # type: ignore

import typer

from jinja_tree.app.config import (
    Config,
)
from jinja_tree.infra.utils import get_config_file_path, read_config_file_or_die

ConfigFileType = Annotated[
    Optional[str],
    typer.Option(
        help="config file path (default: first '.jinja-tree.toml' file found up from current working dir), can also be see with JINJA_TREE_CONFIG_FILE env var",
        envvar="JINJA_TREE_CONFIG_FILE",
    ),
]
LogLevelType = Annotated[
    str, typer.Option(help="log level (DEBUG, INFO, WARNING or ERROR)")
]

ExtraSearchPathsType = Annotated[
    Optional[List[Path]],
    typer.Option(help="Search path to jinja (can be used multiple times)"),
]

AddProcessedFileDirToSearchPathType = Annotated[
    Optional[bool],
    typer.Option(help="add directory of current processed file to jinja search path"),
]

AddRootDirToSearchPathType = Annotated[
    Optional[bool], typer.Option(help="add root directory to jinja search path")
]

AddCwdDirToSearchPathType = Annotated[
    Optional[bool],
    typer.Option(help="add current working directory (CWD) to jinja search path"),
]

ExtensionType = Annotated[
    Optional[List[str]], typer.Option(help="jinja extension to load")
]

ContextPluginsType = Annotated[
    Optional[List[str]],
    typer.Option(
        help="context plugins (full python class path, can be used multiple times)"
    ),
]

StrictUndefinedType = Annotated[
    Optional[bool],
    typer.Option(help="if set, raise an error if a variable does not exist in context"),
]

ActionPluginType = Annotated[
    Optional[List[str]],
    typer.Option(
        help="action plugin (full python class path, can be used multiple times)"
    ),
]

BlankRunType = Annotated[
    bool,
    typer.Option(
        help="if set, execute a blank run (without modifying or deleting anything)"
    ),
]

DisableEmbeddedExtensionsType = Annotated[
    Optional[bool], typer.Option(help="disable embedded jinja extensions")
]

VerboseType = Annotated[
    bool,
    typer.Option(
        help="increase verbosity of the DEBUG log level (note: this forces log-level = DEBUG)"
    ),
]


RootDirType = Annotated[Path, typer.Argument(help="root directory")]


def get_config(
    config_file_path: ConfigFileType = None,
    extra_search_path: ExtraSearchPathsType = None,
    add_cwd_to_search_path: AddCwdDirToSearchPathType = None,
    add_root_dir_to_search_path: AddRootDirToSearchPathType = None,
    strict_undefined: StrictUndefinedType = None,
    jinja_extension: ExtensionType = None,
    disable_embedded_jinja_extensions: DisableEmbeddedExtensionsType = None,
    root_dir: Optional[RootDirType] = None,
    context_plugins: ContextPluginsType = None,
    action_plugins: ActionPluginType = None,
    verbose: VerboseType = False,
    log_level: LogLevelType = "INFO",
) -> Config:
    if not config_file_path:
        config_file_path = get_config_file_path()
    config = read_config_file_or_die(config_file_path)
    if extra_search_path:
        config.extra_search_paths = [str(x) for x in extra_search_path]
    if add_cwd_to_search_path is not None:
        config.add_cwd_to_search_path = add_cwd_to_search_path
    if add_root_dir_to_search_path is not None:
        config.add_root_dir_to_search_path = add_root_dir_to_search_path
    if strict_undefined is not None:
        config.strict_undefined = strict_undefined
    if jinja_extension:
        config.jinja_extensions = jinja_extension
    if disable_embedded_jinja_extensions is not None:
        config.disable_embedded_jinja_extensions = disable_embedded_jinja_extensions
    if root_dir is not None:
        config.root_dir = str(root_dir)
    if context_plugins:
        config.context_plugins = context_plugins
    if action_plugins:
        config.action_plugins = action_plugins
    config.verbose = verbose
    if verbose:
        config.log_level = "DEBUG"
        config.verbose = True
    else:
        config.log_level = log_level
        config.verbose = False
    config.__post_init__()
    return config



from jinja2 import pass_eval_context as eval_context
from jinja2.ext import Extension


@eval_context
def double_quotes(eval_ctx, value):
    return '"' + value + '"'


class DoubleQuotesExtension(Extension):
    """Jinja2 extension to add double quotes around a string.

    Example:
    {{ 'foo-bar'|double_quotes() }}

    => "foo-bar"

    """

    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["double_quotes"] = double_quotes



import fnmatch
import os
import sys
from importlib import import_module
from typing import List, Optional, Type

import dataclasses_json
import stlog
import tomli

from jinja_tree.app.action import ActionPort
from jinja_tree.app.config import (
    Config,
)
from jinja_tree.app.context import ContextPort

SYSTEM_CONFIG_PATH = "/etc/jinja-tree.toml"

logger = stlog.getLogger("jinja_tree")


def import_class_from_string(class_path: str) -> Type:
    module_name, _, class_name = class_path.rpartition(".")
    klass = getattr(import_module(module_name), class_name)
    return klass


def make_context_adapter_from_config(klass, config: Config) -> ContextPort:
    config_name = klass.get_config_name()
    plugin_config = config.context_plugins_configs.get(config_name, {})
    return klass(config, plugin_config)


def make_action_adapter_from_config(klass, config: Config) -> ContextPort:
    config_name = klass.get_config_name()
    plugin_config = config.action_plugins_configs.get(config_name, {})
    return klass(config, plugin_config)


def make_context_adapters_from_config(config: Config) -> List[ContextPort]:
    res: List[ContextPort] = []
    for class_path in config.context_plugins:
        context_adapter_class = import_class_from_string(class_path)
        context_adapter = make_context_adapter_from_config(
            context_adapter_class, config
        )
        if not isinstance(context_adapter, ContextPort):
            raise Exception(
                f"the class pointed by {class_path} does not implement ContextPort interface"
            )
        res.append(context_adapter)
    return res


def make_action_adapters_from_config(config: Config) -> List[ActionPort]:
    res: List[ActionPort] = []
    for class_path in config.action_plugins:
        context_adapter_class = import_class_from_string(class_path)
        action_adapter = make_action_adapter_from_config(context_adapter_class, config)
        if not isinstance(action_adapter, ActionPort):
            raise Exception(
                f"the class pointed by {class_path} does not implement ActionPort interface"
            )
        res.append(action_adapter)
    return res


def get_config_file_path(
    cli_option: Optional[str] = None, cwd: Optional[str] = None
) -> Optional[str]:
    """
    Get the path to the configuration file.

    - if the CLI option is set, it is returned immediately (note: can be set by JINJA_TREE_CONFIG_FILE env var)
    - else, we will try to find it in the current working directory
        or in the parent directory (recursively).

    Args:
        cli_option: The configuration file path provided via CLI
            (if set, the function returns this value immediately)
        cwd: The current working directory (if not set, the current working directory is used)

    Returns:
        The path to the configuration file, or None if not found.

    """
    if cli_option is not None:
        logger.debug(f"config file path read from CLI: {cli_option}")
        return cli_option
    if cwd is None:
        cwd = os.getcwd()
    else:
        cwd = os.path.abspath(cwd)
    path = os.path.join(cwd, ".jinja-tree.toml")
    if os.path.isfile(path):
        logger.debug(f"config file path found in: {path}")
        return os.path.abspath(path)
    parent_path = os.path.dirname(cwd)
    if parent_path == cwd:
        # we are done
        # let's try the system config path?
        if os.path.isfile(SYSTEM_CONFIG_PATH):
            return SYSTEM_CONFIG_PATH
        logger.debug("no config file found")
        return None
    return get_config_file_path(cwd=parent_path)


def is_fnmatch_ignored(key: str, ignores: List[str]) -> bool:
    """
    Check if the given key matches any of the patterns in the ignores list using fnmatch.

    Args:
        key: The key to check.
        ignores: The list of patterns to ignore.

    Returns:
        True if the key matches any of the patterns, False otherwise.
    """
    return any(fnmatch.fnmatch(key, x) for x in ignores)


def setup_logger(log_level: Optional[str] = None):
    if log_level is None:
        log_level = "INFO"
    stlog.setup(level=log_level)


def log_error_and_die(*args, **kwargs):
    setup_logger()  # as we are not that logging is setup at this point
    logger.error(*args, **kwargs)
    sys.exit(1)


def read_config_file_or_die(config_file_path: Optional[str]) -> Config:
    with stlog.LogContext.bind(config_file_path=config_file_path):
        if config_file_path is not None:
            try:
                with open(config_file_path, "rb") as f:
                    data = tomli.load(f)
            except Exception:
                log_error_and_die("cannot read config file", exc_info=True)
            for key in data.keys():
                if key not in ("general", "context", "action"):
                    log_error_and_die(f"invalid section: {key} found in config file")
            general = data.get("general", {})
            try:
                config = Config.from_dict(general)
            except dataclasses_json.undefined.UndefinedParameterError as e:
                log_error_and_die(
                    f"invalid parameters found in general section of the config file: {e.normalized_messages()}"
                )
            config.context_plugins_configs = data.get("context", {})
            config.action_plugins_configs = data.get("action", {})
            try:
                make_context_adapters_from_config(config)
            except dataclasses_json.undefined.UndefinedParameterError as e:
                log_error_and_die(
                    f"invalid parameters found in context section of the config file: {e.normalized_messages()}"
                )
            except Exception:
                log_error_and_die("invalid context plugin configuration", exc_info=True)
            try:
                make_action_adapters_from_config(config)
            except dataclasses_json.undefined.UndefinedParameterError as e:
                log_error_and_die(
                    f"invalid parameters found in action section of the config file: {e.normalized_messages()}"
                )
            except Exception:
                log_error_and_die("invalid action plugin configuration", exc_info=True)
            config.__post_init__()
            return config
    return Config()



import subprocess

from jinja2 import pass_eval_context as eval_context
from jinja2.ext import Extension


@eval_context
def shell(eval_ctx, value, die_on_error=True, encoding="utf8", **kwargs):
    if die_on_error:
        cmd = value
    else:
        cmd = "%s ; exit 0" % value
    return subprocess.check_output(
        cmd, stderr=subprocess.STDOUT, shell=True, encoding=encoding, **kwargs
    )


class ShellExtension(Extension):
    """Jinja2 extension to execute shell commands.

    Example:
    {{ "date"|shell() }}

    => "Thu  7 Nov 2019 15:55:01 CET"

    WARNING: be sure to valid any string submitted to this filter as
             you can open security holes with it

    """

    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["shell"] = shell



import typer
import typer.core

from jinja_tree.app import dump
from jinja_tree.app.action import ActionService
from jinja_tree.app.context import ContextService
from jinja_tree.app.jinja import JinjaService
from jinja_tree.app.jinja_tree import JinjaTreeService
from jinja_tree.infra.controllers.cli_common import (
    ActionPluginType,
    AddCwdDirToSearchPathType,
    AddRootDirToSearchPathType,
    BlankRunType,
    ConfigFileType,
    ContextPluginsType,
    DisableEmbeddedExtensionsType,
    ExtensionType,
    ExtraSearchPathsType,
    LogLevelType,
    RootDirType,
    StrictUndefinedType,
    VerboseType,
    get_config,
)
from jinja_tree.infra.utils import (
    make_action_adapters_from_config,
    make_context_adapters_from_config,
    setup_logger,
)

# disable rich usage in typer
typer.core.rich = None  # type: ignore
app = typer.Typer(add_completion=False)


@app.command()
def tree(
    root_dir: RootDirType,
    config_file: ConfigFileType = None,
    log_level: LogLevelType = "INFO",
    verbose: VerboseType = False,
    extra_search_path: ExtraSearchPathsType = None,
    add_cwd_to_search_path: AddCwdDirToSearchPathType = None,
    add_root_dir_to_search_path: AddRootDirToSearchPathType = None,
    jinja_extension: ExtensionType = None,
    context_plugin: ContextPluginsType = None,
    action_plugin: ActionPluginType = None,
    strict_undefined: StrictUndefinedType = None,
    blank_run: BlankRunType = False,
    disable_embedded_jinja_extensions: DisableEmbeddedExtensionsType = None,
):
    """
    Process a directory tree with the Jinja / Jinja2 templating system.

    """
    config = get_config(
        config_file_path=config_file,
        extra_search_path=extra_search_path,
        add_cwd_to_search_path=add_cwd_to_search_path,
        add_root_dir_to_search_path=add_root_dir_to_search_path,
        jinja_extension=jinja_extension,
        strict_undefined=strict_undefined,
        context_plugins=context_plugin,
        action_plugins=action_plugin,
        disable_embedded_jinja_extensions=disable_embedded_jinja_extensions,
        root_dir=root_dir,
        log_level=log_level,
        verbose=verbose,
    )
    if config.verbose:
        dump("config", config)
    setup_logger(config.log_level)
    context_adapters = make_context_adapters_from_config(config)
    action_adapters = make_action_adapters_from_config(config)
    context_service = ContextService(config=config, adapters=context_adapters)
    action_service = ActionService(config=config, adapters=action_adapters)
    jinja_service = JinjaService(config=config, context_service=context_service)
    jinja_tree_service = JinjaTreeService(
        config=config,
        action_service=action_service,
        jinja_service=jinja_service,
        blank_run=blank_run,
    )
    jinja_tree_service.process()


if __name__ == "__main__":
    app()



import os
import shutil
import sys

import stlog

from jinja_tree.app.action import (
    ActionService,
    IgnoreDirectoryAction,
    IgnoreFileAction,
    ProcessFileAction,
    RenameFileAction,
)
from jinja_tree.app.config import Config
from jinja_tree.app.jinja import JinjaService

IGNORE_FILENAME = ".jinja-tree-ignore"
logger = stlog.getLogger("jinja-tree")


def get_source_content(source_absolute_path: str) -> str:
    with open(source_absolute_path) as f:
        return f.read()


def get_target_content(target_absolute_path: str) -> str:
    with open(target_absolute_path) as f:
        return f.read()


def set_target_content(
    target_absolute_path: str,
    content: str,
    source_absolute_path: str,
    blank_run: bool = False,
):
    if (
        os.path.isfile(target_absolute_path)
        and get_target_content(target_absolute_path) == content
    ):
        logger.debug("No change to apply to the target file", path=target_absolute_path)
        return
    if blank_run:
        print(
            f"[BLANK RUN] Fake setting new content for {target_absolute_path}",
            file=sys.stderr,
        )
        print("[BLANK RUN]<new_content>", file=sys.stderr)
        print(content)
        print("[BLANK RUN]</new_content>", file=sys.stderr)
        return
    logger.info(
        "New content written to target file",
        path=target_absolute_path,
        source_path=source_absolute_path,
    )
    with open(target_absolute_path, "w") as f:
        f.write(content)


def delete_source_file(filepath: str, blank_run: bool = False):
    if blank_run:
        print(f"[BLANK RUN] Fake deleting {filepath}...", file=sys.stderr)
    else:
        logger.info("Deleting a file", path=filepath)
        os.unlink(filepath)


class JinjaTreeService:
    """
    A service class for processing tree of files using Jinja templates.

    Args:
        config: The configuration object.
        action_service: The Action service.
        jinja_service: The JinjaService object to render single strings.
        blank_run: Flag indicating if it is a blank run.
    """

    def __init__(
        self,
        config: Config,
        action_service: ActionService,
        jinja_service: JinjaService,
        blank_run: bool = False,
    ):
        self.config = config
        self.action_service = action_service
        self.jinja_service = jinja_service
        self.blank_run = blank_run

    def trace(self, msg: str, **kwargs):
        if self.config.verbose:
            logger.debug(msg, **kwargs)

    def chdir(self, dirpath: str):
        self.trace("Changing working directory", new_path=dirpath)
        os.chdir(dirpath)

    def do_process_action(self, action: ProcessFileAction):
        if self.config.change_cwd:
            self.chdir(action.dirpath)
        content = get_source_content(action.source_absolute_path)
        logger.debug(
            "Processing a file",
            path=action.source_absolute_path,
            target=action.target_absolute_path,
        )
        result = self.jinja_service.render_string(content, action.source_absolute_path)
        set_target_content(
            action.target_absolute_path,
            result,
            action.source_absolute_path,
            blank_run=self.blank_run,
        )
        if action.delete_original:
            delete_source_file(action.source_absolute_path, blank_run=self.blank_run)

    def do_rename_action(self, action: RenameFileAction):
        if self.blank_run:
            print(
                f"[BLANK RUN] Fake renaming {action.source_absolute_path} to {action.target_absolute_path}...",
                file=sys.stderr,
            )
        else:
            logger.info(
                f"Renaming {action.source_absolute_path} to {action.target_absolute_path}..."
            )
            shutil.move(action.source_absolute_path, action.target_absolute_path)

    def process(self):
        for dirpath, dirnames, filenames in os.walk(self.config.root_dir, topdown=True):
            # Check if the directory should be ignored because of IGNORE_FILENAME
            if IGNORE_FILENAME in filenames:
                self.trace(
                    f"Ignored directory because {IGNORE_FILENAME} found (inside)",
                    path=dirpath,
                )
                # modify in-place, see https://stackoverflow.com/a/19859907
                # to ignore recursively
                dirnames[:] = []
                continue

            # Directory action
            dir_action = self.action_service.get_directory_action(dirpath)
            if isinstance(dir_action, IgnoreDirectoryAction):
                # modify in-place, see https://stackoverflow.com/a/19859907
                # to ignore recursively
                dirnames[:] = []
                continue

            # File actions
            for f in filenames:
                path = os.path.join(dirpath, f)
                action = self.action_service.get_file_action(path)
                if isinstance(action, IgnoreFileAction):
                    pass
                elif isinstance(action, ProcessFileAction):
                    try:
                        self.do_process_action(action)
                    except Exception:
                        logger.error(
                            "An error occurred while processing the file",
                            path=action.source_absolute_path,
                        )
                        raise
                elif isinstance(action, RenameFileAction):
                    self.do_rename_action(action)
                else:
                    raise Exception("unknown action type: %s", action)



import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Dict, List

import stlog

from jinja_tree.app.config import Config

logger = stlog.getLogger("jinja-tree")


@dataclass
class FileAction:
    """Abstract base class for file actions.

    Attributes:
        source_absolute_path: absolute path for the source file (the template).
    """

    source_absolute_path: str

    @property
    def dirpath(self) -> str:
        return os.path.dirname(self.source_absolute_path)


@dataclass
class IgnoreFileAction(FileAction):
    """This is a concrete implementation of FileAction to represent a "do nothing with this file" action."""

    pass


@dataclass
class ProcessFileAction(FileAction):
    """This is a concrete implementation of FileAction to represent a "process this file with jinja" action.

    Attributes:
        target_absolute_path: absolute path for the target file (the rendered file).
        delete_original: if True, the original file will be deleted after the rendering.
    """

    target_absolute_path: str
    delete_original: bool = False


@dataclass
class RenameFileAction(FileAction):
    """This is a concrete implementation of fileAction to represent a "rename this file" action."""

    target_absolute_path: str


@dataclass
class DirectoryAction:
    """Abstract base class for directory actions.

    Attributes:
        source_absolute_path: absolute path for the source directory.
    """

    source_absolute_path: str


@dataclass
class IgnoreDirectoryAction(DirectoryAction):
    """This is a concrete implementation of DirectoryAction to represent a "do nothing with this directory" action.

    All files in this directory or subdirectories will be ignored (recursively).
    """

    pass


@dataclass
class BrowseDirectoryAction(DirectoryAction):
    """This is a concrete implementation of DirectoryAction to represent a "browse this directory" action.

    The directory itself won't be changed but all files and subdirectories will be scanned for actions.
    """

    pass


CONCRETE_DIRECTORY_ACTIONS = [IgnoreDirectoryAction, BrowseDirectoryAction]
DEFAULT_CONCRETE_DIRECTORY_ACTION = BrowseDirectoryAction
CONCRETE_FILE_ACTIONS = [IgnoreFileAction, ProcessFileAction, RenameFileAction]
DEFAULT_CONCRETE_FILE_ACTION = IgnoreFileAction


class ActionPort(ABC):
    """This is the abstract interface for FileActionPort adapters."""

    @abstractmethod
    def __init__(self, config: Config, plugin_config: Dict[str, Any]):
        """
        Construct a new FileActionPort object given a configuration object
        and a plugin configuration dict.

        Args:
            config (Config): The configuration object.
            plugin_config (Dict[str, Any]): The plugin configuration dict.

        """
        pass

    @classmethod
    @abstractmethod
    def get_config_name(cls) -> str:
        """
        Return the name of the configuration object.

        For example, if we return "foo", it means that the configuration in the TOML file
        is located under:

        [action.foo]
        # ... some configuration ...

        Returns:
            The name of the configuration object.

        """
        pass

    @abstractmethod
    def get_file_action(self, absolute_path: str) -> FileAction:
        """Return the action to execute on the file at the given absolute path.

        Note:
        - absolute_path is checked to be a file before calling this method.

        Attributes:
            absolute_path: absolute path for the file to process.
        """
        pass

    @abstractmethod
    def get_directory_action(self, absolute_path: str) -> DirectoryAction:
        """Return the action to execute on the directory at the given absolute path.

        Note:
        - absolute_path is checked to be a directory before calling this method.

        Attributes:
            absolute_path: absolute path for the directory to process.
        """
        pass


class ActionService:
    """This service is responsible for returning the action to execute on the file at the given absolute path.

    It's a kind of proxy object on FileActionPort adapters to factorize some common logic.

    Attributes:
        config: The configuration object.
        adapter: The FileActionPort adapter to use.

    """

    def __init__(self, config: Config, adapters: List[ActionPort]):
        self.config = config
        self.adapters = adapters

    def get_file_action(self, absolute_path: str) -> FileAction:
        """Return the action to execute on the file at the given absolute path.

        Attributes:
            absolute_path: absolute path for the file to process.
        """
        assert os.path.isfile(absolute_path)
        if not self.adapters:
            raise Exception("no action plugins configured!")
        for adapter in self.adapters:
            a = adapter.get_file_action(absolute_path)
            if not isinstance(a, tuple(CONCRETE_FILE_ACTIONS)):
                raise Exception(f"bad action: {a} returned by action plugin: {adapter}")
            if isinstance(a, DEFAULT_CONCRETE_FILE_ACTION):
                # => IgnoreFileAction, let's try next adapter
                continue
            target_absolute_path: str = a.target_absolute_path  # type: ignore
            if os.path.exists(target_absolute_path) and not os.path.isfile(
                target_absolute_path
            ):
                logger.warning(
                    f"target file: {target_absolute_path} is not a file => ignoring"
                )
                return IgnoreFileAction(source_absolute_path=absolute_path)
            return a
        return DEFAULT_CONCRETE_FILE_ACTION(source_absolute_path=absolute_path)

    def get_directory_action(self, absolute_path: str) -> DirectoryAction:
        """Return the action to execute on the directory at the given absolute path.

        Attributes:
            absolute_path: absolute path for the directory to process.
        """
        assert os.path.isdir(absolute_path)
        for adapter in self.adapters:
            a = adapter.get_directory_action(absolute_path)
            if not isinstance(a, tuple(CONCRETE_DIRECTORY_ACTIONS)):
                raise Exception(f"bad action: {a} returned by action plugin: {adapter}")
            if isinstance(a, DEFAULT_CONCRETE_DIRECTORY_ACTION):
                # => BrowseDirectoryAction, let's try next adapter
                continue
            return a
        return DEFAULT_CONCRETE_DIRECTORY_ACTION(source_absolute_path=absolute_path)



import json
import sys

RICH_AVAILABLE = True
try:
    from rich import print as rprint
except ImportError:
    rprint = print  # type: ignore
    RICH_AVAILABLE = False


def dump(name: str, obj):
    rprint(f"<{name} dump>", file=sys.stderr)
    if RICH_AVAILABLE:
        rprint(obj, file=sys.stderr)
    else:
        rprint(
            json.dumps(
                obj, indent=4, sort_keys=True, default=lambda o: "<not serializable>"
            ),
            file=sys.stderr,
        )
    rprint(f"</{name} dump>", file=sys.stderr)



import os
from dataclasses import dataclass, field
from typing import Any, Dict, List

from dataclasses_json import DataClassJsonMixin, Undefined
from dotenv import dotenv_values

from jinja_tree.app.config import Config
from jinja_tree.app.context import ContextPort
from jinja_tree.infra.utils import is_fnmatch_ignored

ENV_CONTEXT_ADAPTER_DEFAULT_IGNORES: List[str] = []
DOTENV_CONTEXT_ADAPTER_DEFAULT_PATH = ".env"
DOTENV_CONTEXT_ADAPTER_DEFAULT_IGNORES: List[str] = []


@dataclass
class EnvContextConfig(DataClassJsonMixin):
    ignores: List[str] = field(
        default_factory=lambda: list(ENV_CONTEXT_ADAPTER_DEFAULT_IGNORES)
    )
    dataclass_json_config = {"undefined": Undefined.RAISE}  # noqa: RUF012


class EnvContextAdapter(ContextPort):
    def __init__(self, config: Config, plugin_config: Dict[str, Any]):
        self.config = config
        self.plugin_config = EnvContextConfig.from_dict(plugin_config)

    @classmethod
    def get_config_name(cls) -> str:
        return "env"

    def get_context(self) -> Dict[str, Any]:
        if self.plugin_config.ignores == ["*"]:
            return {}
        return {
            x: y
            for x, y in os.environ.items()
            if not is_fnmatch_ignored(x, self.plugin_config.ignores)
        }


@dataclass
class DotEnvContextConfig(DataClassJsonMixin):
    path: str = DOTENV_CONTEXT_ADAPTER_DEFAULT_PATH
    ignores: List[str] = field(
        default_factory=lambda: list(DOTENV_CONTEXT_ADAPTER_DEFAULT_IGNORES)
    )
    dataclass_json_config = {"undefined": Undefined.RAISE}  # noqa: RUF012

    def __post_init__(self):
        self.path = os.path.abspath(self.path)


class DotEnvContextAdapter(ContextPort):
    def __init__(self, config: Config, plugin_config: Dict[str, Any]):
        self.config = config
        self.plugin_config = DotEnvContextConfig.from_dict(plugin_config)

    @classmethod
    def get_config_name(cls) -> str:
        return "dotenv"

    def get_context(self) -> Dict[str, Any]:
        if not self.plugin_config.path:
            return {}
        if not os.path.isfile(self.plugin_config.path):
            return {}
        if self.plugin_config.ignores == ["*"]:
            return {}
        return {
            x: y
            for x, y in dotenv_values(self.plugin_config.path).items()
            if not is_fnmatch_ignored(x, self.plugin_config.ignores)
        }


class ConfigurationContextAdapter(ContextPort):
    def __init__(self, config: Config, plugin_config: Dict[str, Any]):
        self.config = config
        self.plugin_config = plugin_config

    @classmethod
    def get_config_name(cls) -> str:
        return "config"

    def get_context(self) -> Dict[str, Any]:
        return self.plugin_config



import sys

import typer
import typer.core

from jinja_tree.app import dump
from jinja_tree.app.context import ContextService
from jinja_tree.app.jinja import JinjaService
from jinja_tree.infra.controllers.cli_common import (
    AddCwdDirToSearchPathType,
    ConfigFileType,
    ContextPluginsType,
    DisableEmbeddedExtensionsType,
    ExtensionType,
    ExtraSearchPathsType,
    LogLevelType,
    StrictUndefinedType,
    VerboseType,
    get_config,
)
from jinja_tree.infra.utils import make_context_adapters_from_config, setup_logger

# disable rich usage in typer
typer.core.rich = None  # type: ignore
app = typer.Typer(add_completion=False)


@app.command()
def pipe(
    config_file: ConfigFileType = None,
    log_level: LogLevelType = "INFO",
    verbose: VerboseType = False,
    extra_search_path: ExtraSearchPathsType = None,
    add_cwd_to_search_path: AddCwdDirToSearchPathType = None,
    jinja_extension: ExtensionType = None,
    context_plugin: ContextPluginsType = None,
    strict_undefined: StrictUndefinedType = None,
    disable_embedded_jinja_extensions: DisableEmbeddedExtensionsType = None,
):
    """
    Process the standard input with Jinja templating system and return the result on the standard output.

    """
    config = get_config(
        config_file_path=config_file,
        extra_search_path=extra_search_path,
        context_plugins=context_plugin,
        add_cwd_to_search_path=add_cwd_to_search_path,
        jinja_extension=jinja_extension,
        strict_undefined=strict_undefined,
        disable_embedded_jinja_extensions=disable_embedded_jinja_extensions,
        log_level=log_level,
        verbose=verbose,
    )
    setup_logger(config.log_level)
    if config.verbose:
        dump("config", config)
    context_adapters = make_context_adapters_from_config(config)
    context_service = ContextService(config=config, adapters=context_adapters)
    jinja_service = JinjaService(config=config, context_service=context_service)
    print(jinja_service.render_string(sys.stdin.read()))


if __name__ == "__main__":
    app()



import json

from jinja2 import pass_eval_context as eval_context
from jinja2.ext import Extension


@eval_context
def from_json(eval_ctx, value):
    return json.loads(value)


class FromJsonExtension(Extension):
    """Jinja2 extension to load a JSON string into a Python object."""

    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["from_json"] = from_json



import os
from dataclasses import dataclass, field
from typing import Any, Dict, List

from dataclasses_json import DataClassJsonMixin, Undefined

EMBEDDED_EXTENSIONS = [
    "jinja_tree.app.embedded_extensions.from_json.FromJsonExtension",
    "jinja_tree.app.embedded_extensions.from_toml.FromTomlExtension",
    "jinja_tree.app.embedded_extensions.shell.ShellExtension",
    "jinja_tree.app.embedded_extensions.fnmatch.FnMatchExtension",
    "jinja_tree.app.embedded_extensions.double_quotes.DoubleQuotesExtension",
    "jinja_tree.app.embedded_extensions.urlencode.UrlEncodeExtension",
]

JINJA_TREE_URL = "https://github.com/fabien-marty/jinja-tree"


def make_default_context_plugins() -> List[str]:
    return [
        "jinja_tree.infra.adapters.context.ConfigurationContextAdapter",
        "jinja_tree.infra.adapters.context.EnvContextAdapter",
        "jinja_tree.infra.adapters.context.DotEnvContextAdapter",
    ]


def make_default_action_plugins() -> List[str]:
    return ["jinja_tree.infra.adapters.action.ExtensionsActionAdapter"]


@dataclass
class Config(DataClassJsonMixin):
    # Global config
    extra_search_paths: List[str] = field(default_factory=list)
    add_root_dir_to_search_path: bool = True
    add_cwd_to_search_path: bool = True
    add_processed_file_dir_to_search_path: bool = False
    change_cwd: bool = True
    jinja_extensions: List[str] = field(default_factory=list)
    strict_undefined: bool = True
    root_dir: str = field(default_factory=os.getcwd)
    disable_embedded_jinja_extensions: bool = False
    verbose: bool = False
    log_level: str = "INFO"
    context_generated_comment_line1: str = "*** GENERATED FILE - DO NOT EDIT ***"
    context_generated_comment_line2: str = (
        "This file was generated by jinja-tree (%s) from the template file: {{relative_filepath}}"
        % JINJA_TREE_URL
    )
    context_plugins: List[str] = field(default_factory=make_default_context_plugins)
    action_plugins: List[str] = field(default_factory=make_default_action_plugins)

    context_plugins_configs: Dict[str, Dict[str, Any]] = field(
        init=False, default_factory=dict
    )
    action_plugins_configs: Dict[str, Dict[str, Any]] = field(
        init=False, default_factory=dict
    )

    dataclass_json_config = {"undefined": Undefined.RAISE}  # noqa: RUF012

    @property
    def resolved_extensions(self) -> List[str]:
        if self.disable_embedded_jinja_extensions:
            return self.jinja_extensions
        return EMBEDDED_EXTENSIONS + self.jinja_extensions

    def __post_init__(self):
        # replaces paths by absolute paths
        self.root_dir = os.path.abspath(self.root_dir)
        self.extra_search_paths = [os.path.abspath(x) for x in self.extra_search_paths]



import os
from typing import Optional

import stlog
from jinja2 import (
    ChoiceLoader,
    DictLoader,
    Environment,
    FileSystemLoader,
    StrictUndefined,
    Undefined,
    UndefinedError,
)

from jinja_tree.app.config import Config
from jinja_tree.app.context import ContextService

logger = stlog.getLogger("jinja-tree")


class JinjaService:
    """Service to render a string with Jinja2 calling the context service to get the context."""

    def __init__(
        self,
        config: Config,
        context_service: ContextService,
    ):
        self.context_service = context_service
        self.template_name = "__content"
        self.config = config

    def render_string(self, content: str, absolute_path: Optional[str] = None) -> str:
        """
        Renders the given content as a Jinja template string.

        The context is retrieved from the ContextService.

        Args:
            content: The content to render as a template.
            absolute_path: The absolute path of the template file (if available).

        Returns:
            str: The rendered output as a string.
        """
        loader1 = DictLoader({self.template_name: content})
        paths = []
        if self.config.add_root_dir_to_search_path:
            paths.append(self.config.root_dir)
        if self.config.add_processed_file_dir_to_search_path:
            paths.append(os.getcwd())
        if self.config.extra_search_paths:
            paths = paths + [str(x) for x in self.config.extra_search_paths]
        loader2 = FileSystemLoader(paths, followlinks=True)
        loader = ChoiceLoader([loader1, loader2])
        env = Environment(
            loader=loader,
            undefined=StrictUndefined if self.config.strict_undefined else Undefined,
            extensions=self.config.resolved_extensions,
        )
        template = env.get_template(self.template_name)
        template.globals = self.context_service.get_context(absolute_path)
        try:
            output = template.render()
        except UndefinedError:
            logger.error(
                "error during the template rendering because of undefined variables (you can use the --no-strict-undefined flag to disable this)",
                path=absolute_path,
            )
            raise
        return output



import fnmatch

from jinja2 import pass_eval_context as eval_context
from jinja2.ext import Extension


@eval_context
def _fnmatch(eval_ctx, value, pattern):
    return fnmatch.fnmatch(value, pattern)


class FnMatchExtension(Extension):
    """Jinja2 extension to provide a fnmatch filter.

    Example:
    {{ 'foo-bar'|fnmatch('foo-*') }}

    => True

    """

    def __init__(self, environment):
        super().__init__(environment)
        environment.filters["fnmatch"] = _fnmatch
